#!/usr/bin/python

#
#  Copyright (C) 2004 Red Hat, Inc.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

import ldap

# FIXME:
#  Searches like (& (objectClass=statelessSnapshot)(statelessSnapshotName=Blah))
#  are lame because we should just be doing a BASE search for
#  statelessSnapshotName=Blah,statelessConfigurationName=Foo,dc=stateless,ou=services,...

def escape_attribute(unescaped):
    result = ""
    for ch in unescaped:
        if ch in ',+\"\\<>;':
            result = result + '\\'
        result = result + ch
    if result[0] in ' #':
        result = "\\" + result
    #FIXME: handle trailing space and hash

    return result

def search_sync(ldap_obj, dn, scope, filterstr='(objectClass=*)'):
    result = []
    obj_list = ldap_obj.search_s(dn, scope, filterstr)
    for i in obj_list:
        result.append(make_entry_model_from_search(i))
    return result

# For when you only expect to get one entry back (or zero):
def search_sync_individual(ldap_obj, dn, scope, filterstr='(objectClass=*)'):
    entry_list = search_sync(ldap_obj, dn, scope, filterstr)    
    if len(entry_list)>0:
        return entry_list[0]
    else:
        return None

# (Assumes that the parent_dn has been escaped)
def get_child_dn (parent_dn, rdn_name, escaped_rdn_value):
    return "%s=%s, %s"%(rdn_name, escaped_rdn_value, parent_dn)

class LDAPConfigMissing(Exception):
    def __init__(self, *args):
        Exception.__init__ (self, *args)

class Directory:
    def __init__ (self, cfg):
        self.uri = cfg.get_ldap_uri ()
        if self.uri == None:
            raise LDAPConfigMissing, "No LDAP_SERVER defined in /etc/sysconfig/stateless"
        
        self.base_dn = cfg["LDAP_BASE_DN"]
        if self.base_dn == None:
            raise LDAPConfigMissing, "No LDAP_BASE_DN defined in /etc/sysconfig/stateless"

    def make_ldap_obj (self):
        raise NotImplementedError

    def is_writable (self):
        raise NotImplementedError

class ReadOnlyDirectory (Directory):
    def __init__ (self, cfg):
        Directory.__init__ (self, cfg)

    def make_ldap_obj (self):
        return ldap.ldapobject.SimpleLDAPObject (self.uri)

    def is_writable(self):
        return False

class ReadWriteDirectory (Directory):
    def __init__ (self, cfg):
        self.admin_dn = cfg["LDAP_ADMIN_DN"]
        if self.admin_dn == None:
            raise LDAPConfigMissing, "No LDAP_ADMIN_DN defined in /etc/sysconfig/stateless"
        
        self.admin_passwd = cfg["LDAP_ADMIN_PASSWD"]
        if self.admin_passwd == None:
            raise LDAPConfigMissing, "No LDAP_ADMIN_PASSWD defined in /etc/sysconfig/stateless"
        
        Directory.__init__ (self, cfg)

    def make_ldap_obj (self):
        ldap_obj = ldap.ldapobject.SimpleLDAPObject(self.uri)
        ldap_obj.simple_bind_s(self.admin_dn, self.admin_passwd)
        
        return ldap_obj

    def is_writable (self):
        return True

def make_entry_model_from_search(search_result):
    (dn, attr_dict) = search_result    
    return EntryModel(dn, attr_dict)

# A representation of an LDAP entry; attributes and DN are stored in memory; getting at related entries is done by talking to the server.
# the attributes are stored as a dictionary from names to lists of string values
class EntryModel:
    def __init__(self, dn, attr_dict):
        self.dn = dn
        self.attr_dict = attr_dict

    def __str__(self):
        return "EntryModel(DN:\"%s\", %s)"%(self.dn, self.attr_dict)
        
    def get_attr_value_list(self, attr_name):
        try:
            return self.attr_dict[attr_name]
        except:
            return []

    def get_attr_value(self, attr_name):
        attr_list = self.get_attr_value_list(attr_name)
        try:
            return attr_list[0]
        except:
            return None

    # Create a child model in memory (without actually doing anything to the server)
    def make_child_model(self, rdn_attr, rdn_value, objectClass, other_attributes):
        full_attributes = {rdn_attr:[rdn_value], "objectClass":[objectClass]}
        for (key,value) in other_attributes.iteritems():            
            full_attributes[key]=value
        return EntryModel(get_child_dn (self.dn, rdn_attr, rdn_value), full_attributes)

    # Modifying the directory:
    def add_sync(self, ldap_obj):
        modlist = []
        for (key, valuelist) in self.attr_dict.iteritems():
            if valuelist!=[]:
                modlist.append((key,valuelist))
        ldap_obj.add_s(self.dn, modlist)

    def ensure_exists(self, ldap_obj):
        try:
            self.add_sync(ldap_obj)
        except ldap.ALREADY_EXISTS:
            pass

    def update_entry(self, ldap_obj):
        modlist = []
        for (key, valuelist) in self.attr_dict.iteritems():
            modlist.append((ldap.MOD_REPLACE,key,valuelist))
        ldap_obj.modify_s(self.dn, modlist)

    def delete_entry(self, ldap_obj, recurse):
        if recurse:
            children = search_sync(ldap_obj, self.dn, ldap.SCOPE_ONELEVEL)
            for child in children:
                child.delete_entry(ldap_obj, True)
        
        ldap_obj.delete_s(self.dn)

class Provisioning:
    def __init__(self, config, config_name, protocol, snapshot_name):
        self.config        = config
        self.config_name   = config_name
        self.protocol      = protocol
        self.snapshot_name = snapshot_name

    def __str__(self):
        return "(config \"%s\", protocol \"%s\", snapshot \"%s\")" % (self.config_name, self.protocol, self.snapshot_name)

    def is_valid(self):
        return self.config and self.config_name and self.snapshot_name

class StatelessService:
    def __init__ (self, directory):
        self.directory = directory
        self.ldap_obj = self.directory.make_ldap_obj()
        self.root_dc = search_sync_individual(self.ldap_obj, self.directory.base_dn, ldap.SCOPE_BASE)

        if self.directory.is_writable():
            self.ensure_exists()

        def get_child(ldap_obj, dn, rdn):
            child_dn = "%s, %s" % (rdn, dn)
            return search_sync_individual(ldap_obj, child_dn, ldap.SCOPE_BASE)

        self.stateless_entry = get_child(self.ldap_obj, self.root_dc.dn, "dc=stateless,ou=services")
        self.hosts_entry     = get_child(self.ldap_obj, self.stateless_entry.dn, "dc=hosts")
        self.servers_entry   = get_child(self.ldap_obj, self.stateless_entry.dn, "dc=servers")

    def get_ldap_server_uri (self):
        return self.directory.uri

    def get_ldap_base_dn (self):
        return self.directory.base_dn
        
    def ensure_exists(self):
        baseEntryModel = EntryModel(self.directory.base_dn, {})
        
        services = baseEntryModel.make_child_model("ou", "services", "organizationalUnit", {})
        services.ensure_exists(self.ldap_obj)
        
        statelessService = services.make_child_model("dc", "stateless", "statelessContainer", {})
        statelessService.ensure_exists(self.ldap_obj)

        hosts = statelessService.make_child_model("dc", "hosts", "statelessContainer", {})
        hosts.ensure_exists(self.ldap_obj)
        
        servers = statelessService.make_child_model("dc", "servers", "statelessContainer", {})
        servers.ensure_exists(self.ldap_obj)

    def get_configuration_list(self):
        filterstr = '(objectClass=statelessConfiguration)'
        return search_sync(self.ldap_obj, self.stateless_entry.dn, ldap.SCOPE_ONELEVEL, filterstr)

    def get_configuration(self, config_name):
        filterstr = '(& (objectClass=statelessConfiguration)(statelessConfigurationName=%s))' % (config_name)
        return search_sync_individual(self.ldap_obj, self.stateless_entry.dn, ldap.SCOPE_ONELEVEL, filterstr)

    def get_server_list(self):
        filterstr = '(objectClass=statelessServer)'
        return search_sync(self.ldap_obj, self.servers_entry.dn, ldap.SCOPE_ONELEVEL, filterstr)
    
    def get_server(self, server_name):
        filterstr = '(& (ObjectClass=statelessServer)(cn=%s))' % (server_name)
        return search_sync_individual(self.ldap_obj, self.servers_entry.dn, ldap.SCOPE_ONELEVEL, filterstr)

    def get_client_list(self):
        filterstr = '(objectClass=statelessClient)'
        return search_sync(self.ldap_obj, self.hosts_entry.dn, ldap.SCOPE_ONELEVEL, filterstr)
    
    def get_client(self, mac_address):
        filterstr = '(& (objectClass=statelessClient)(macAddress=%s))' % (mac_address)
        return search_sync_individual(self.ldap_obj, self.hosts_entry.dn, ldap.SCOPE_ONELEVEL, filterstr)

    # return a Provisioning object giving which config name/version should be being run by the specified client
    def get_client_provisioning(self, mac_address):
        filterstr = "(& (objectClass=statelessClient)(macAddress=%s))" % mac_address
        client = search_sync_individual(self.ldap_obj, self.hosts_entry.dn, ldap.SCOPE_ONELEVEL, filterstr)
        if client == None:
            return None
        
        config_name = client.get_attr_value("statelessConfigurationName")
        if config_name == None:
            return None
        
        filterstr = "(& (objectClass=statelessConfiguration)(statelessConfigurationName=%s))" % config_name
        config = search_sync_individual(self.ldap_obj, self.stateless_entry.dn, ldap.SCOPE_ONELEVEL, filterstr)
        if config == None:
            return None

        snapshot_name = client.get_attr_value("statelessSnapshotName")
        if snapshot_name == None:
            snapshot_name = config.get_attr_value("statelessDefaultSnapshotName")

        return Provisioning(config, config_name, None, snapshot_name)

    def get_snapshot(self, config_name, snapshot_name):
        filterstr = "(& (objectClass=statelessSnapshot)(statelessSnapshotName=%s))" % (snapshot_name)
        config_dn = get_child_dn (self.stateless_entry.dn, "statelessConfigurationName", config_name)
        return search_sync_individual(self.ldap_obj, config_dn, ldap.SCOPE_ONELEVEL, filterstr)
    
    def get_snapshot_list(self, config_name):
        filterstr = '(objectClass=statelessSnapshot)'
        config_dn = get_child_dn (self.stateless_entry.dn, "statelessConfigurationName", config_name)
        return search_sync(self.ldap_obj, config_dn, ldap.SCOPE_ONELEVEL, filterstr)

    def get_snapshot_location(self, config_name, snapshot_name, protocol):
        snapshot = self.get_snapshot(config_name, snapshot_name)
        if snapshot == None:
            return (None, None)

        server_names = snapshot.get_attr_value_list("statelessSnapshotServer")
        if len(server_names) == 0:
            return (None, None)

        if protocol == "rsync":
            attr_name = "statelessRsyncPath"
        else:
            assert protocol == "nfs"
            attr_name = "statelessNfsPath"

        for server_name in server_names:
            server = self.get_server(server_name)
            if server == None:
                continue

            snapshots_path = server.get_attr_value(attr_name)
            if snapshots_path != None:
                return (server_name, snapshots_path)

        return (None, None)

    def get_client_status(self, mac_address):
        client = self.get_client(mac_address)
        if client == None:
            return None

        return search_sync_individual(self.ldap_obj, client.dn, ldap.SCOPE_ONELEVEL, '(&(dc=status))')

    def get_reported_client_provisioning(self, mac_address):
        status = self.get_client_status(mac_address)
        if status == None:
            return None

        protocol      = status.get_attr_value("statelessProtocolName")
        config_name   = status.get_attr_value("statelessConfigurationName")
        snapshot_name = status.get_attr_value("statelessSnapshotName")

        return Provisioning(self.get_configuration(config_name), config_name, protocol, snapshot_name)

    def add_configuration(self, name, default_snapshot):
        attrs = { "statelessDefaultSnapshotName" : [] }
        if default_snapshot != None:
            attrs["statelessDefaultSnapshotName"].append(default_snapshot)
        
        config = self.stateless_entry.make_child_model("statelessConfigurationName", name, "statelessConfiguration", attrs)
        config.add_sync(self.ldap_obj)
        return config
        
    def add_server(self, name, nfspath=None, rsyncpath=None):
        attrs = { "statelessNfsPath" : [] }
        if nfspath != None:
            attrs["statelessNfsPath"] = [nfspath]

        attrs["statelessRsyncPath"] = []
        if rsyncpath != None:
            attrs["statelessRsyncPath"] = [rsyncpath]
            
        server = self.servers_entry.make_child_model("cn", name, "statelessServer", attrs)
        server.add_sync(self.ldap_obj)
        return server

    def add_client(self, mac_address, config_name, snapshot_name=None):
        attrs = { "statelessConfigurationName" : [] }
        if config_name != None:
            attrs["statelessConfigurationName"].append(config_name)
        
        attrs["statelessSnapshotName"] = []
        if snapshot_name != None:
            attrs["statelessSnapshotName"].append(snapshot_name)
        
        client = self.hosts_entry.make_child_model("macAddress", mac_address, "statelessClient", attrs)
        client.add_sync(self.ldap_obj)
        return client

    def add_snapshot(self, config_name, snapshot_name, servers):
        config = self.get_configuration(config_name)
        if config == None:
            return None
        
        attrs = { "statelessSnapshotServer" : servers }
        snapshot = config.make_child_model("statelessSnapshotName", snapshot_name, "statelessSnapshot", attrs)
        snapshot.add_sync(self.ldap_obj)
        return snapshot

    def add_client_status(self, mac_address, protocol, config_name, snapshot_name):
        client = self.get_client(mac_address)
        if client == None:
            return None

        attrs = { "statelessProtocolName" : [] }
        if protocol != None:
            attrs["statelessProtocolName"].append(protocol)

        attrs["statelessConfigurationName"] = []
        if config_name != None:
            attrs["statelessConfigurationName"].append(config_name)

        attrs["statelessSnapshotName"] = []
        if snapshot_name != None:
            attrs["statelessSnapshotName"].append(snapshot_name)
            
        status = client.make_child_model("dc", "status", "statelessClientStatus", attrs)
        status.add_sync(self.ldap_obj)

    def update_client(self, mac_address, config_name, snapshot_name):
        client = self.get_client(mac_address)
        if client != None:
            client.attr_dict["statelessConfigurationName"] = []
            if config_name != None:
                client.attr_dict["statelessConfigurationName"].append(config_name)
               
            client.attr_dict["statelessSnapshotName"] = []
            if snapshot_name != None:
                client.attr_dict["statelessSnapshotName"].append(snapshot_name)
            
            client.update_entry(self.ldap_obj)

    def update_client_status(self, mac_address, protocol, config_name, snapshot_name):
        status = self.get_client_status(mac_address)
        if status != None:
            status.attr_dict["statelessProtocolName"]      = []
            if protocol != None:
                status.attr_dict["statelessProtocolName"].append(protocol)
            
            status.attr_dict["statelessConfigurationName"] = []
            if config_name != None:
                status.attr_dict["statelessConfigurationName"].append(config_name)
            
            status.attr_dict["statelessSnapshotName"]      = []
            if snapshot_name != None:
                status.attr_dict["statelessSnapshotName"].append(snapshot_name)
            
            status.update_entry(self.ldap_obj)
        
    def update_server(self, server_name, nfs_path, rsync_path):
        server = self.get_server(server_name)
        if server != None:
            server.attr_dict["statelessNfsPath"] = []
            if nfs_path != None:
                server.attr_dict["statelessNfsPath"].append(nfs_path)
            
            server.attr_dict["statelessRsyncPath"] = rsync_path
            if rsync_path != None:
                server.attr_dict["statelessRsyncPath"].append(rsync_path)
            
            server.update_entry(self.ldap_obj)

    def set_default_snapshot(self, config_name, snapshot_name):
        config = self.get_configuration(config_name)
        if config != None:
            config.attr_dict["statelessDefaultSnapshotName"] = [snapshot_name]
            config.update_entry(self.ldap_obj)

    def delete_server(self, server_name):
        server = self.get_server(server_name)
        if server != None:
            server.delete_entry(self.ldap_obj, True)
            
    def delete_client(self, mac_address):
        client = self.get_client(mac_address)
        if client != None:
            client.delete_entry(self.ldap_obj, True)
