#!/usr/bin/python

#
#  Copyright (C) 2004 Red Hat, Inc.
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

# This module includes a uility function for reinstalling grub
# after rsync'ing to one of the Stateless Linux
# boot partitions.
#
# Notes:
#
#   * Grub will be installed on the MBR of the drive where the /boot
#     partition lives, other configurations aren't supported
#   * All the grub files must already be in /boot, and the grub.conf
#     file must look approximately like what anaconda writes out
#   * This file uses various internal functions from booty; it
#     may be quite fragile against booty upgrades. It has been
#     tested with booty-0.38-1 from FC2.
#   * This doesn't handle swapping the / partition; the grub.conf
#     has to use root=LABEL=/ (which doesn't work for raid?)
#
# Usage:
#
# statelessg = statelessgrub.statelessGrub(bootdev, bootmount, instRoot = "/")
# statelessg.reinstall()
#
# @bootdev: Device name (e.g., /dev/hda1) of device for /boot
#          (must be a separate partition just for /boot,
#          mountpoint must end in /boot)
# @bootmount: Place where bootdev is currently mounted
# @instRoot: Installed image root (defaults to "/")
#

import os
import re
import sys
import rhpl.executil

sys.path += ['/usr/lib/anaconda', '/usr/lib/booty']

import booty
import bootloaderInfo
import butil
import isys

# create an initial grub.conf file from the files in bootmount/boot.
# This is a quick hack to get something going; Ideally we'd be able to
# use grubby in the future.  We'll need better grubby support though.

grub_template = """#
# grub.conf generated by statelessgrub.py
#

default=%(default)s
timeout=10
splashimage=%(drive)s/grub/splash.xpm.gz
"""

kernel_template = """
title %(title)s
	root %(drive)s
        kernel %(kernel)s ro root=LABEL=/ selinux=0 rhgb quiet
        initrd %(initrd)s
"""

class statelessGrub:
    def __init__ (self, bootdev, rootmount):
        self.bootdev = bootdev

        if (rootmount[-1:] == '/'):
            self.rootmount = rootmount[:-1]
        else:
            self.rootmount = rootmount

        self.bootmount = self.rootmount + "/boot"
        self.bl = booty.getBootloader()

    def reinstall (self):
        if os.path.isfile(self.grubconf_path()):
            self.munge_grub_conf()
        else:
            self.write_grub_conf()

        self.copy_grub_files()
        self.write_device_map()
            
        bootpart = self.get_partition_name (self.bootdev)
        grubpath = bootpart + "/grub"
        root = self.get_disk_name (self.bootdev)
        cmd = "root %s\ninstall %s/stage1 d %s %s/stage2 p %s/grub.conf" % \
              (bootpart, grubpath, root, grubpath, grubpath)

        # install the bootloader
        bootloaderInfo.syncDataToDisk(self.bootdev[5:], self.bootmount, self.rootmount)
        p = os.pipe()
        os.write(p[1], cmd + '\n')
        os.close(p[1])
        rhpl.executil.execWithRedirect('/sbin/grub' ,
                                [ "grub",  "--batch", "--no-floppy",
                                   "--device-map=%s/device.map" % (self.bootmount + "/grub") ],
                                   stdin = p[0],
                                   root = self.rootmount)
        os.close(p[0])

    def write_device_map (self):
        bootpart = self.get_partition_name (self.bootdev)
        devfile = self.bootmount + "/grub/device.map"
        if os.path.isfile(devfile):
            return

        f = open(devfile, "w")
        f.write("# this device map was generated by stateless\n")
        f.write("(fd0)     /dev/fd0\n")
        f.write("%s        %s\n" % (bootpart, self.bootdev))

    def copy_grub_files (self):
        # copy the stage files over into /boot
        rhpl.executil.execWithRedirect( "/sbin/grub-install",
                                  ["/sbin/grub-install", "--just-copy" ],
                                   root = self.rootmount)
        
    def munge_grub_conf (self):
        bootpart = self.get_partition_name (self.bootdev)
        conffile = self.grubconf_path()
        old = open (conffile, "r")
        contents = old.read (-1)
        old.close()
        contents = re.sub ("root \([^\)]*\)", "root %s" % bootpart, contents)
        contents = re.sub ("splashimage=\([^\)]*\)", "splashimage=%s" % bootpart, contents)
        new = open (conffile, "w")
        new.write (contents)
        new.close ()

    def write_grub_conf (self):
        kernelList = self.find_kernels()
        bootpart = self.get_partition_name (self.bootdev)

        default_index = -1
        i = 0
        for (blah, blah, blah, default) in kernelList:
            if default:
                default_index = i
            i = i + 1
        assert default_index != -1
        
        new = open (self.grubconf_path(), "w")
        new.write (grub_template % {
            "drive": bootpart,
            "default": default_index
        })
        for (label, longlabel, version, default) in kernelList:
            new.write (kernel_template % {
                "drive": bootpart,
                "kernel": "/vmlinuz-%s" % version,
                "initrd": "/initrd-%s.img" % version,
                "title":  longlabel,
                })
        
    def grubconf_path (self):
        return self.bootmount + "/grub/grub.conf"

    def get_partition_name (self, devname):
        if devname[:5] != "/dev/":
            raise ValueError("Device name doesn't start with /dev/")
        return self.bl.grubbyPartitionName(devname[5:])

    def get_disk_name (self, devname):
        if devname[:5] != "/dev/":
            raise ValueError("Device name doesn't start with /dev/")
        (disk, blah) = bootloaderInfo.getDiskPart(devname[5:])
        return "(" + self.bl.grubbyDiskName(disk) + ")"

    def find_kernels (self):
        kernels = {}
        up = 1
        smp = 2

        default_up_version = None
        default_smp_version = None
        
        for file in os.listdir (self.bootmount):
            m = re.match ("vmlinuz-(.*)", file)
            if m:
                version = m.group(1)
                is_smp = 0
                m = re.match ("(.*)smp", version)
                if m:
                    version = m.group(1)
                    is_smp = 1
                if not kernels.has_key(version):
                    kernels[version] = 0

                if is_smp:
                    kernels[version] = kernels[version] | smp
                    default_smp_version = version
                else:
                    kernels[version] = kernels[version] | up
                    default_up_version = version

	use_smp = default_smp_version != None and (isys.smpAvailable() or isys.htavailable())

        kernelList = []
        defaultLongLabel = butil.getProductName()
        for version in kernels.keys():
            if (kernels[version] == smp | up):
                kernelList.append (("linux", defaultLongLabel, version + "smp",
                                    use_smp and version == default_smp_version))
                kernelList.append (("linux-up", defaultLongLabel + "-up", version,
                                    (not use_smp) and version == default_up_version))
            elif (kernels[version] == smp):
                kernelList.append (("linux", defaultLongLabel, version + "smp",
                                    use_smp and version == default_smp_version))
            else:
                kernelList.append (("linux", defaultLongLabel, version,
                                    (not use_smp) and version == default_up_version))
        
        return kernelList
